#ifdef HAVE_IMLIB
# include <Imlib.h>
#endif
#ifdef HAVE_IMLIB2
# include <X11/Xlib.h>
# include <Imlib2.h>
#endif
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "grab.h"


/* Define this to indicate compiling for an X windows supported
 * platform.
 */
#define HAVE_X


#ifdef HAVE_IMLIB
/* Handle used for passing to all Imlib functions, not needed for
 * Imlib2. This is foundationally declared in main.c.
 */
extern void *imlib_handle;
#endif  /* HAVE_IMLIB */


/* Pointer to last grab error message. */
static const gchar *last_grab_error = NULL;

/* Grab callback data structure. */
typedef struct {

#define GRAB_UNDETERMINED	0	/* Not yet determined. */
#define GRAB_WINDOW		1	/* Grab a window. */
#define GRAB_RECTANGULAR	2	/* Grab rectangular area. */
#define GRAB_ABORT		5	/* User aborted. */
	gint grab_type;

	GdkWindow *root_win;	/* Pointer to root window, read only. */

	gbool active;		/* Indicates if a grab is active, if this
				 * is FALSE then no grab stuff should be
				 * processed on the signal callbacks.
				 */

	gint gtk_main_level;	/* Relative to our GTK+ main level, not
				 * global GTK+ main level.
				 */

	gint grab_button_number;	/* Indicates which button is
					 * currently pressed, can be -1
					 * for no button pressed.
					 */

	/* Coordinates for grab, x[0] and y[0] represent starting
	 * coordinates for grab_type GRAB_RECTANGULAR or GRAB_WINDOW.
	 *
	 * The x[1] and y[1] are only defined if grab_type is
	 * GRAB_RECTANGULAR.
	 *
	 * All these values are defined if grab_type is GRAB_UNDETERMINED.
	 */
	gint x[2], y[2];

#ifdef HAVE_X
        Display *xdisplay;
	int xscreen_num;
        Window xrootwin;
        GC gc_inverse;
	/* Last drawn rectangle position and size. */
	Bool have_last_rect;
	int last_rect_x, last_rect_y;
	unsigned int last_rect_width, last_rect_height;
#endif

} grab_data_struct;


#ifdef HAVE_X
static void draw_X_grab_rectangle(grab_data_struct *grab_data, Bool draw_new);
static gint grab_X_window(
	imgview_struct *iv, grab_data_struct *grab_data,
	gint *width, gint *height,
	gbool include_decorations, gbool resize_on_load
);
static gint grab_X_rectangular(
        imgview_struct *iv, grab_data_struct *grab_data,
        gint *width, gint *height,
        gbool include_decorations, gbool resize_on_load
);
#endif

static gint grab_delay_tocb(gpointer data);
static gint grab_motion_cb(
        GtkWidget *widget, GdkEventMotion *motion, gpointer data
);
static gint grab_button_cb(
        GtkWidget *widget, GdkEventButton *button, gpointer data
);

const gchar *iv_grab_get_error(void);
gint iv_grab_start(
	imgview_struct *iv,
	gint *width, gint *height,
	glong delay,			/* In milliseconds. */
	gbool include_decorations,
        gbool resize_on_load,
	GtkWidget *grab_window,
	GdkCursor *grab_cursor
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))



#ifdef HAVE_X
/*
 *	Draws an X grab rectangle.
 *
 *	This will use the inverse GC on the given grab_data structure to;
 *	first reset the bits drawn by a previous call to this function
 *	(if there was one, check grab_data->have_last_rect == True) by
 *	drawing the rectangle bits using the previous geometry values on
 *	the grab_data structure, then draw the new rectangle bits if
 *	draw_new is True (using the same inverse GC).
 *
 *	Note that if draw_new is False then grab_data->have_last_rect will
 *	be set to False, indicating for the next call to this function
 *	there would have been no previous rectangle drawn.
 */
static void draw_X_grab_rectangle(grab_data_struct *grab_data, Bool draw_new)
{
	static gbool reenterent = FALSE;
	Display *dpy;
	Window xrootwin;
	Drawable d;
	int x, y;
	int width, height;
	GC gc;


	if(grab_data == NULL)
	    return;

	dpy = grab_data->xdisplay;
	xrootwin = grab_data->xrootwin;
	gc = grab_data->gc_inverse;
	if((dpy == NULL) || (xrootwin == None) || (gc == None))
	    return;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;

	/* Get the root window as the drawable to draw to. */
	d = (Drawable)xrootwin;

	/* Get new rectangle bits geometry values. */
	x = MIN(grab_data->x[0], grab_data->x[1]);
        y = MIN(grab_data->y[0], grab_data->y[1]);
        width = ABSOLUTE(grab_data->x[1] - grab_data->x[0]);
        height = ABSOLUTE(grab_data->y[1] - grab_data->y[0]);

	/* Draw prev/last/old rectangle? This is to effectivly reset the
	 * bits drawn by a prior call to this function.
	 */
	if(grab_data->have_last_rect)
	    XDrawRectangle(
		dpy, d, gc,
		grab_data->last_rect_x,
		grab_data->last_rect_y,
		grab_data->last_rect_width,
		grab_data->last_rect_height
	    );

	/* Draw the new rectangle bits and the new rectangle has
	 * valid size?
	 */
	if(draw_new && (width > 0) && (height > 0))
	{
	    XDrawRectangle(dpy, d, gc, x, y, width, height);

	    /* Since the new rectangle bits have been drawn, its geometry
	     * values should be recorded so that the next call to this
	     * function resets the bits just drawn.
	     */
            grab_data->have_last_rect = True;
            grab_data->last_rect_x = x;
            grab_data->last_rect_y = y;
            grab_data->last_rect_width = width;
            grab_data->last_rect_height = height;
        }
	else
	{
	    /* Do not draw new rectangle bits. */

	    /* Need to mark that there would be no prior rectangle
	     * drawn for the next call to this function.
	     */
	    grab_data->have_last_rect = False;
	}

	reenterent = FALSE;
}
#endif	/* HAVE_X */
#ifdef HAVE_X
/*
 *      Processes a window grab procedure, the pointer will be warped
 *	(moved) to the coordinates specified in grab_data structure and
 *	then the pointer will be queried to see which Window the
 *	pointer is over. That Window (or next child if include_decorations
 *	is FALSE) will have its geometry fetched and a rectangular area
 *	will be grabbed from the root window that matches the geometry
 *	of the Window in question.
 *
 *      Returns 0 on success or non-zero on error.
 */
static gint grab_X_window(
	imgview_struct *iv, grab_data_struct *grab_data,
	gint *width, gint *height,
	gbool include_decorations,
	gbool resize_on_load
)
{
	static gbool reenterent = FALSE;
	int x, y;
	Display *dpy;
	Window xrootwin, xwin;
	Window xrootrtn;		/* Dummy root for fetches. */
	int x_return, y_return, xr2, yr2;
	unsigned int	width_return, height_return,
			border_width_return,
			depth_return;
	unsigned int mask;
	gint status;
	gint cx, cy, cwidth, cheight, rwidth, rheight;
#ifdef HAVE_IMLIB
        ImlibImage *imlib_image = NULL;
#endif


	if((iv == NULL) || (grab_data == NULL))
	    return(-1);

	dpy = grab_data->xdisplay;
	xrootwin = grab_data->xrootwin;
	if((dpy == NULL) || (xrootwin == None))
	    return(-1);

	if(reenterent)
	{
	    last_grab_error = "Another grab operation is already in progress";
	    return(-3);
	}
	else
	    reenterent = FALSE;


	/* Make sure output has been flushed before grabbing, this
	 * is to ensure we don't have any messy marks that we drew
	 * in the image that we want to grab.
	 */
	XFlush(dpy);

	/* Grab X server so things don't change while getting 
	 * window geometry.
	 */
	XGrabServer(dpy);


	/* Get grab window coordinates relative to root window. */
	x = (int)grab_data->x[0];
	y = (int)grab_data->y[0];

	/* Find child of root window that the given coordinates lie
	 * under.
	 */
	XWarpPointer(dpy, None, xrootwin, 0, 0, 0, 0, x, y);
        if(!XQueryPointer(
            dpy, xrootwin,
            &xrootrtn, &xwin, &x_return, &y_return, &xr2, &yr2, &mask
        ))
	{
            last_grab_error =
		"XQueryPointer() failed to find clicked on Window";
            XUngrabServer(dpy);
            reenterent = FALSE;
	    return(-1);
	}

	/* If coordinates do not lie within a child window, then the
	 * window to grab is the xrootwin itself.
	 */
	if(xwin == None)
	    xwin = xrootwin;


        /* At this point we have the X id of the root window and the
         * immediate child or the root window that the pointer was over.
         *
         * We need to handle two possible cases, if the root and child
         * window are the same, then we grab the root window. Otherwise
         * we grab the portion of the root window that the child window is
         * displayed at.
         *
         * In either case the values for variables x, y, rwidth, rheight,
         * cwidth, and cheight will be set.
         */
        if(xwin == xrootwin)
        {
            /* Grabbing the root window, get size of root window. */
            XGetGeometry(
                dpy, xwin, &xrootrtn,
                &x_return, &y_return, &width_return, &height_return,
                &border_width_return, &depth_return
            );
            cx = (gint)x_return;
            cy = (gint)y_return;
            rwidth = cwidth = (gint)width_return;
            rheight = cheight = (gint)height_return;
        }
        else
        {
            /* Grabbing a child window, there are two posibilites now.
             * The first is that we just grab the child window without
             * the WM decorations. The second is that we grab the
             * entire child window and its WM decorations if any.
             */

            /* First get size of root window. */
            XGetGeometry(
                dpy, xrootwin, &xrootrtn,
                &x_return, &y_return, &width_return, &height_return,
                &border_width_return, &depth_return
            );
            rwidth = (gint)width_return;
            rheight = (gint)height_return;

            /* Now work on getting the clicked on child window. */
            if(include_decorations)
            {
                /* Grab child window with its WM decorations or the
                 * child window has no WM decorations.
                 */
                XGetGeometry(
                    dpy, xwin, &xrootrtn,
                    &x_return, &y_return, &width_return, &height_return,
                    &border_width_return, &depth_return
                );
                cx = (gint)x_return;
                cy = (gint)y_return;
                cwidth = (gint)width_return;
                cheight = (gint)height_return;
            }
	    else
            {
                /* Do not grab WM decorations and this window is
                 * suspected to have decorations.
                 */
                Window tw;

                /* Get position of child window. */
                XGetGeometry(
                    dpy, xwin, &xrootrtn,
                    &x_return, &y_return, &width_return, &height_return,
                    &border_width_return, &depth_return
                );
                cx = (gint)x_return;
                cy = (gint)y_return;
		cwidth = (gint)width_return;
		cheight = (gint)height_return;

                /* Now get the next child if any. */
		if(!XQueryPointer(
		    dpy, xwin,
		    &xrootrtn, &tw, &x_return, &y_return, &xr2, &yr2, &mask
		))
		{
		    last_grab_error =
			"XQueryPointer() failed to find undecorated clicked on Window";
		    XUngrabServer(dpy);
		    reenterent = FALSE;
		    return(-1);
		}
                if(tw != None)
		{
		    /* Got next child, so update xwin and its geometry. */
                    xwin = tw;
		    XGetGeometry(
			dpy, xwin, &xrootrtn,
			&x_return, &y_return, &width_return, &height_return,
			&border_width_return, &depth_return
		    );
		    cx += (gint)x_return;
		    cy += (gint)y_return;
		    cwidth = (gint)width_return;
		    cheight = (gint)height_return;
		}
	    }

	}
        /* At this point we now have the size of the root window in
         * rwidth and rheight plus the child window geometry relative
         * to the root window at x, y, cwidth, and cheight.
         */
        /* Sanitize values of child window geometry. */
        if((cx + cwidth) > rwidth)
            cwidth = rwidth - cx;
        if(cx < 0)
        {
            cwidth += cx;
            cx = 0;
        }
        if((cy + cheight) > rheight)
            cheight = rheight - cy;
        if(cy < 0)
        {
            cheight += cy;
            cy = 0;
        }
        /* Check if child window still has area. */
        if((cwidth < 1) || (cheight < 1))
        {
            last_grab_error =
		"Window has no area (width and/or height < 1)";
            XUngrabServer(dpy);
	    reenterent = FALSE;
            return(-1);
        }

	/* We now have the rectangular area we want to grab, specified in
	 * cx, cy, cwidth, and cheight.
	 */

	/* Reset status to -1, indicating error. If we were able to get
	 * the image then we'll set it to 0.
	 */
	status = -1;

        /* Begin getting image. */
#ifdef HAVE_IMLIB
        imlib_image = Imlib_create_image_from_drawable(
            imlib_handle, xrootwin,
            None, cx, cy, cwidth, cheight
        );
        if(imlib_image != NULL)
        {
            const guint8 *src_data = (const guint8 *)imlib_image->rgb_data;
	    GtkWidget *w, *w2;


            /* Need to realize changes. */
            Imlib_changed_image(imlib_handle, imlib_image);


            /* Update child window size with size of the Imlib image. */
            cwidth = imlib_image->rgb_width;
            cheight = imlib_image->rgb_height;


	    /* Get image viewer toplevel widget if it is a GtkWindow. */
	    if(ImgViewToplevelIsWindow(iv))
		w = ImgViewGetToplevelWidget(iv);
	    else
		w = NULL;
	    /* Get image viewer view widget. */
	    w2 = (GtkWidget *)ImgViewGetViewWidget(iv);

            /* Resize image viewer? */
            if((w2 != NULL) && resize_on_load)
            {
                gtk_widget_set_usize(
                    w2, cwidth, cheight
                );
                gtk_widget_queue_resize(w2);
                /* Notify toplevel to resize. */
                if(w != NULL)
                    gtk_widget_queue_resize(w);
            }

            /* Load new image to image viewer, any currently loaded
             * image will be unloaded first.
             */
            ImgViewLoad(
                iv,
                imlib_image->rgb_width, imlib_image->rgb_height,
                0,              /* Auto calculate bpl. */
                IMGVIEW_FORMAT_RGB,     /* Use RGB since Imlib image. */
                src_data
            );

            /* Destroy Imlib image, it is no longer needed in the
             * cache.
             */
            Imlib_kill_image(imlib_handle, imlib_image);
            imlib_image = NULL;

	    /* Mark success in grabbing image. */
	    status = 0;
        }
        else
        {
            last_grab_error = "Imlib failed to get X Drawable";
        }
#endif	/* HAVE_IMLIB */

	/* Ungrab X server. */
	XUngrabServer(dpy);

	/* Update size return. */
	if(width != NULL)
	    (*width) = cwidth;
	if(height != NULL)
	    (*height) = cheight;

	reenterent = FALSE;
	return(status);
}
#endif	/* HAVE_X */
#ifdef HAVE_X
/*
 *      Processes a rectangular grab procedure, the geometry of the
 *	rectangle will be obtained from the given grab_data structure.
 *	This geometry will then be used to get a rectangular image from
 *	the root window.
 *
 *      Returns 0 on success or non-zero on error.
 */
static gint grab_X_rectangular(
        imgview_struct *iv, grab_data_struct *grab_data,
        gint *width, gint *height,
        gbool include_decorations,
        gbool resize_on_load
)
{
        static gbool reenterent = FALSE;
        Display *dpy;
        Window xrootwin;
        Window xrootrtn;                /* Dummy root for fetches. */
        int x_return, y_return;
        unsigned int    width_return, height_return,
                        border_width_return,
                        depth_return;
        gint status;
        gint cx, cy, cwidth, cheight, rwidth, rheight;
#ifdef HAVE_IMLIB
        ImlibImage *imlib_image = NULL;
#endif


        if((iv == NULL) || (grab_data == NULL))
            return(-1);

        dpy = grab_data->xdisplay;
        xrootwin = grab_data->xrootwin;
        if((dpy == NULL) || (xrootwin == None))
            return(-1);

        if(reenterent)
        {
            last_grab_error = "Another grab operation is already in progress";
            return(-3);
        }
        else
            reenterent = FALSE;


        /* Make sure output has been flushed before grabbing, this
         * is to ensure we don't have any messy marks that we drew
         * in the image that we want to grab.
         */
        XFlush(dpy);

        /* Grab X server so things don't change while getting
         * window geometry.
         */
        XGrabServer(dpy);


	/* Get size of root window. */
	XGetGeometry(
	    dpy, xrootwin, &xrootrtn,
	    &x_return, &y_return, &width_return, &height_return,
	    &border_width_return, &depth_return
	);
	rwidth = (gint)width_return;
	rheight = (gint)height_return;

	/* Get size of child area from grab_data. */
	cx = MIN(grab_data->x[0], grab_data->x[1]);
        cy = MIN(grab_data->y[0], grab_data->y[1]);
	cwidth = ABSOLUTE(grab_data->x[1] - grab_data->x[0]);
        cheight = ABSOLUTE(grab_data->y[1] - grab_data->y[0]);

        /* At this point we now have the size of the root window in
         * rwidth and rheight plus the child window geometry relative
         * to the root window at x, y, cwidth, and cheight.
         */
        /* Sanitize values of child window geometry. */
        if((cx + cwidth) > rwidth)
            cwidth = rwidth - cx;
        if(cx < 0)
        {
            cwidth += cx;
            cx = 0;
        }
        if((cy + cheight) > rheight)
            cheight = rheight - cy;
        if(cy < 0)
        {
            cheight += cy;
            cy = 0;
        }
        /* Check if child window still has area. */
        if((cwidth < 1) || (cheight < 1))
        {
            last_grab_error =
		"Grab rectangle has no area (width and/or height < 1)";
            XUngrabServer(dpy);
            reenterent = FALSE;
            return(-1);
        }

        /* We now have the rectangular area we want to grab, specified in
         * cx, cy, cwidth, and cheight.
         */

        /* Reset status to -1, indicating error. If we were able to get
         * the image then we'll set it to 0.
         */
        status = -1;

        /* Begin getting image. */
#ifdef HAVE_IMLIB
        imlib_image = Imlib_create_image_from_drawable(
            imlib_handle, xrootwin,
            None, cx, cy, cwidth, cheight
        );
        if(imlib_image != NULL)
        {
            const guint8 *src_data = (const guint8 *)imlib_image->rgb_data;
            GtkWidget *w, *w2;


            /* Need to realize changes. */
            Imlib_changed_image(imlib_handle, imlib_image);


            /* Update child window size with size of the Imlib image. */
            cwidth = imlib_image->rgb_width;
            cheight = imlib_image->rgb_height;


            /* Get image viewer toplevel widget if it is a GtkWindow. */
            if(ImgViewToplevelIsWindow(iv))
                w = ImgViewGetToplevelWidget(iv);
            else
                w = NULL;
            /* Get image viewer view widget. */
            w2 = (GtkWidget *)ImgViewGetViewWidget(iv);

            /* Resize image viewer? */
            if((w2 != NULL) && resize_on_load)
            {
                gtk_widget_set_usize(
                    w2, cwidth, cheight
                );
                gtk_widget_queue_resize(w2);
                /* Notify toplevel to resize. */
                if(w != NULL)
                    gtk_widget_queue_resize(w);
            }

            /* Load new image to image viewer, any currently loaded
             * image will be unloaded first.
             */
            ImgViewLoad(
                iv,
                imlib_image->rgb_width, imlib_image->rgb_height,
                0,              /* Auto calculate bpl. */
                IMGVIEW_FORMAT_RGB,     /* Use RGB since Imlib image. */
                src_data
            );

            /* Destroy Imlib image, it is no longer needed in the
             * cache.
             */
            Imlib_kill_image(imlib_handle, imlib_image);
            imlib_image = NULL;

            /* Mark success in grabbing image. */
            status = 0;
        }
        else
        {
            last_grab_error = "Imlib failed to get X Drawable";
        }
#endif  /* HAVE_IMLIB */

        /* Ungrab X server. */
        XUngrabServer(dpy);

        /* Update size return. */
        if(width != NULL)
            (*width) = cwidth;
        if(height != NULL)
            (*height) = cheight;

        reenterent = FALSE;
        return(status);
}
#endif	/* HAVE_X */


/*
 *	Grab delay timeout callback, this is called when the delay
 *	time expires in iv_grab_start() and to break out of the GTK+ main
 *	loop set just after the setting of this callback.
 */
static gint grab_delay_tocb(gpointer data)
{
	gint *main_level = (gint *)data;

	if((*main_level) > 0)
	{
	    (*main_level) = (*main_level) - 1;
	    gtk_main_quit();
	}

	return(FALSE);		/* Don't call this timeout again. */
}

/*
 *	Grab motion event callback.
 */
static gint grab_motion_cb(
        GtkWidget *widget, GdkEventMotion *motion, gpointer data
)
{
        grab_data_struct *grab_data = (grab_data_struct *)data;
        if((motion == NULL) || (grab_data == NULL))
            return(FALSE);

        /* Do not process event if grabbing is not schedualed to
         * be active yet.
         */
	if(!grab_data->active)
	    return(TRUE);

	/* Button pressed? */
	if(grab_data->grab_button_number > -1)
	{
	    switch(grab_data->grab_type)
	    {
	      case GRAB_WINDOW:
		/* Do nothing. */
		break;

	      case GRAB_RECTANGULAR:
		/* Update grab coordinates. */
                grab_data->x[1] = (gint)motion->x;
                grab_data->y[1] = (gint)motion->y;
#ifdef HAVE_X
		draw_X_grab_rectangle(grab_data, True);
#endif
		break;

	      case GRAB_ABORT:
                /* Do nothing. */
		break;
	    }
	}

	return(TRUE);
}

/*
 *	Grab button callback.
 */
static gint grab_button_cb(
        GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	grab_data_struct *grab_data = (grab_data_struct *)data;
	if((button == NULL) || (grab_data == NULL))
	    return(FALSE);

	/* Do not process event if grabbing is not schedualed to
	 * be active yet.
	 */
	if(!grab_data->active)
            return(TRUE);

	/* Button press? */
	if(button->type == GDK_BUTTON_PRESS)
	{
	    /* Get number of the button that was pressed. */
	    gint button_num = button->button;

	    /* Update pressed button number on grab_data structure. */
	    grab_data->grab_button_number = button_num;

	    /* Set grab_type by which button number was pressed. */
	    switch(button_num)
	    {
	      case 1:	/* Button1. */
		grab_data->grab_type = GRAB_WINDOW;
		grab_data->x[0] = (gint)button->x;
                grab_data->y[0] = (gint)button->y;
		break;

              case 2:	/* Button2. */
                grab_data->grab_type = GRAB_RECTANGULAR;
                grab_data->x[0] = (gint)button->x;
                grab_data->y[0] = (gint)button->y;
		grab_data->x[1] = grab_data->x[0];
		grab_data->y[1] = grab_data->y[0];
                break;

	      default:	/* All other button numbers, assume abort. */
		grab_data->grab_type = GRAB_ABORT;
		break;
	    }
	}
	else
	{
	    /* Assume button release. */

            /* Mark on the grab_data structure that there is no button
	     * that is currently pressed.
	     */
            grab_data->grab_button_number = -1;


	    /* If grabbing a rectangular area, then update rectangle
	     * end coordinates and redraw rectangle outline.
	     */
	    if(grab_data->grab_type == GRAB_RECTANGULAR)
	    {
		grab_data->x[1] = (gint)button->x;
                grab_data->y[1] = (gint)button->y;
#ifdef HAVE_X
		draw_X_grab_rectangle(grab_data, False);
#endif
	    }

	    /* Break out of our GTK+ main loop when a button is
	     * released and grab is marked active.
	     */
	    if(grab_data->gtk_main_level > 0)
	    {
		grab_data->gtk_main_level--;
		gtk_main_quit();
	    }
	}

	return(TRUE);
}


/*
 *	Returns pointer to last error message (or NULL if there wasn't
 *	any). The returned pointer must not be deallocated.
 */
const gchar *iv_grab_get_error(void)
{
	return(last_grab_error);
}

/*
 *	Start a grab procedure.
 *
 *	The given grab_window must be a valid and mapped GtkWindow, it
 *	will be used to grab the pointer and report relative events to.
 *
 *	The pointer will be grabbed to query which window or rectangular
 *	area to grab. On success the given image viewer will have its
 *	image data cleared and then replaced with the grabbed image.
 *
 *	Return values:
 *
 *		0	Success
 *		-1	General error (see iv_grab_get_error())
 *		-2	User did not select grab window/area properly
 *		-3	System error
 *		-4	User canceled
 *
 *	When 0 is returned, the exact error message can be retrieved by
 *	calling iv_grab_get_error().
 */
gint iv_grab_start(
        imgview_struct *iv,
        gint *width, gint *height,
	glong delay,			/* In milliseconds. */
        gbool include_decorations,
        gbool resize_on_load,
        GtkWidget *grab_window,
	GdkCursor *grab_cursor
)
{
	static gbool reenterent = FALSE;
	gbool pointer_grab_called = FALSE;
	gint i, status;
	static gint main_level;	/* gtk_main() level counter. */
	GdkWindow *window;
	GdkWindow *root_win = (GdkWindow *)GDK_ROOT_PARENT();
	GtkWidget *w;
	grab_data_struct *grab_data;
	guint grab_signal_id[3];
#ifdef HAVE_X
	Display *dpy;
	int scr_num;
	Window xrootwin;
	XGCValues gcv;
	GC gc_inverse;
#endif


	/* Check if reentering. */
	if(reenterent)
	{
	    last_grab_error = "A grab is already in progress";
	    return(-1);
	}
	else
	{
	    reenterent = TRUE;
	}

	/* Reset last grab error string. */
        last_grab_error = NULL;

	/* Reset return size. */
        if(width != NULL)
            (*width) = 0;
        if(height != NULL)
            (*height) = 0;

#ifdef HAVE_IMLIB
        if(imlib_handle == NULL)
        {
            last_grab_error = "Imlib not initialized";
	    reenterent = FALSE;
            return(-3);
        }
#endif

        if(iv == NULL)
        {
            last_grab_error = "Image viewer structure is NULL";
	    reenterent = FALSE;
            return(-1);
        }
        if((grab_window != NULL) ?
	    GTK_WIDGET_NO_WINDOW(grab_window) : 1
	)
        {
            last_grab_error = "Grab window is NULL";
            reenterent = FALSE;
            return(-1);
        }


	/* Wait for delay. */
	main_level = 0;
	if(delay > 0)
	{
	    main_level++;
	    gtk_timeout_add(
		delay,
		(GtkFunction)grab_delay_tocb,
		(gpointer)&main_level
	    );
	    gtk_main();
	}
	while(main_level > 0)
	{
	    main_level--;
	    gtk_main_quit();
	}


        /* Root window must be valid. */
        if(root_win == NULL)
        {
            last_grab_error =
		"GDK_ROOT_PARENT(): Unable to obtain root rindow";
            reenterent = FALSE;
            return(-1);
        }

	/* Get GdkWindow from grab_window. */
	window = grab_window->window;
	if(window == NULL)
	{
            last_grab_error = "Grab window is NULL";
            reenterent = FALSE;
            return(-1);
	}

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

#ifdef HAVE_X
	/* Get X resources. */
        dpy = GDK_WINDOW_XDISPLAY(root_win);
	scr_num = DefaultScreen(dpy);
        xrootwin = GDK_WINDOW_XWINDOW(root_win);
        if((dpy == NULL) || (xrootwin == None))
        {
            last_grab_error = "Cannot obtain X resources";
            reenterent = FALSE;
            return(-3);
        }

        /* Sync X to ensure most recent characteristics are being
         * displayed just before grabbing.
         */
        XSync(dpy, False);
#endif	/* HAVE_X */

	/* Allocate grab callback data. */
	grab_data = (grab_data_struct *)g_malloc0(sizeof(grab_data_struct));
	if(grab_data == NULL)
	{
            last_grab_error = "Memory allocation error";
            reenterent = FALSE;
            return(-3);
	}
	/* Reset grab data structure values. */
	grab_data->grab_type = GRAB_UNDETERMINED;
	grab_data->active = FALSE;
	grab_data->root_win = root_win;
	grab_data->gtk_main_level = 0;
	grab_data->grab_button_number = -1;
#ifdef HAVE_X
	grab_data->xdisplay = dpy;
	grab_data->xscreen_num = scr_num;
	grab_data->xrootwin = xrootwin;
	/* Create GCs for grab related drawing. */
	gcv.function = GXinvert;
	gcv.plane_mask = AllPlanes;
	gcv.foreground = WhitePixel(dpy, scr_num);
	gcv.background = BlackPixel(dpy, scr_num);
	gcv.line_width = 1;
	gcv.line_style = LineSolid;
	gcv.cap_style = CapNotLast;
        gcv.join_style = JoinMiter;
        gcv.fill_style = FillSolid;
        gcv.fill_rule = EvenOddRule;
	gcv.subwindow_mode = IncludeInferiors;
	grab_data->gc_inverse = gc_inverse = XCreateGC(
	    dpy, xrootwin,
	    GCFunction | GCForeground | GCBackground | GCPlaneMask |
	    GCLineWidth | GCLineStyle | GCCapStyle | GCJoinStyle |
	    GCFillStyle | GCFillRule | GCSubwindowMode,
	    &gcv
	);
	grab_data->have_last_rect = False;
#endif

	/* Install tempory signal handlers for the grab window. */
	grab_signal_id[0] = gtk_signal_connect(
	    GTK_OBJECT(grab_window), "button_press_event",
	    GTK_SIGNAL_FUNC(grab_button_cb), grab_data
	);
        grab_signal_id[1] = gtk_signal_connect(
            GTK_OBJECT(grab_window), "button_release_event",
            GTK_SIGNAL_FUNC(grab_button_cb), grab_data
        );
        grab_signal_id[2] = gtk_signal_connect(
            GTK_OBJECT(grab_window), "motion_notify_event",
            GTK_SIGNAL_FUNC(grab_motion_cb), grab_data
        );


/* Macro to ungrab the pointer as needed, checks and updates 
 * pointer_grab_called.
 */
#define DO_UNGRAB_POINTER	\
{ \
 if(pointer_grab_called) \
 { \
  gdk_flush(); \
  gdk_pointer_ungrab(GDK_CURRENT_TIME); \
  gdk_flush(); \
  pointer_grab_called = FALSE; \
 } \
}

/* Macro to deallocate grab_data. */
#ifdef HAVE_X
#define DO_FREE_GRAB_CBDATA	\
{ \
 if(grab_data != NULL) \
 { \
  /* Deallocate X resources (only if HAVE_X is #defined). */ \
  if(grab_data->gc_inverse != None) \
  { \
   XFreeGC(dpy, grab_data->gc_inverse); \
   grab_data->gc_inverse = None; \
  } \
\
  grab_data->xdisplay = NULL; \
  grab_data->xscreen_num = 0; \
  grab_data->xrootwin = None; \
\
\
  /* Deallocate grab structure itself. */ \
  g_free(grab_data); \
  grab_data = NULL; \
 } \
}
#else
#define DO_FREE_GRAB_CBDATA	\
{ \
 if(grab_data != NULL) \
 { \
\
\
\
  /* Deallocate grab structure itself. */ \
  g_free(grab_data); \
  grab_data = NULL; \
 } \
}
#endif

/* Macro to ungrab pointer, disconnect signal handlers, and deallocate
 * grab callback data structure.
 */
#define GRAB_CLEANUP	\
{ \
 /* Ungrab pointer as needed. */ \
 DO_UNGRAB_POINTER \
\
 /* Disconnect our tempory signal handlers. */ \
 for(i = 0; i < 3; i++) \
 { \
  gtk_signal_disconnect(GTK_OBJECT(grab_window), grab_signal_id[i]); \
  grab_signal_id[i] = (guint)-1; \
 } \
\
 /* Break out of any remaining GTK+ main levels. */ \
 while(grab_data->gtk_main_level > 0) \
 { \
  grab_data->gtk_main_level--; \
  gtk_main_quit(); \
 } \
\
\
 /* Deallocate grab callback data structure. */ \
 DO_FREE_GRAB_CBDATA \
}

	/* Begin grabbing the pointer so grab related events get reported
	 * to and with respect to the given grab window.
	 */
	gdk_flush();	/* Must flush GDK output before grabbing pointer. */
	status = gdk_pointer_grab(
	    window,
	    FALSE,
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_POINTER_MOTION_MASK,
	    /* root_win */ NULL,
	    grab_cursor,
	    GDK_CURRENT_TIME
	);
        gdk_flush();    /* Must flush GDK output after grabbing pointer. */
	/* Was there an error while attempting to grab the pointer? */
	switch(status)
	{
	  case GrabInvalidTime:
            last_grab_error =
                "Invalid time specified for grabbing pointer";
          case GrabNotViewable:
            last_grab_error =
                "Grab window is not viewable to grab pointer";
          case GrabFrozen:
            last_grab_error =
                "Grab is frozen and unable to grab pointer";
	    /* Do not continue if grab pointer failed, return and indicate
	     * systems error.
	     */
	    GRAB_CLEANUP
            reenterent = FALSE;
            return(-3);
            break;

          default:
	    /* Note gdk_pointer_grab() was called so we know to ungrab
	     * it later.
	     */
	    pointer_grab_called = TRUE;
            break;
        }

        /* Mark grab as now being active on the grab_data structure,
	 * this will allow grab related events to be processed in the
	 * signal callbacks we set above.
         */
        grab_data->active = TRUE;

	/* Enter our GTK+ main loop. */
	grab_data->gtk_main_level++;
	gtk_main();

	/* Mark grab as expired, so no more grab related events will be
	 * processed on our signal callbacks.
	 */
	grab_data->active = FALSE;

	/* Ungrab the pointer, we do not need it anymore. This will also
	 * mark pointer_grab_called as FALSE so it won't be ungrabed
	 * again.
	 */
	DO_UNGRAB_POINTER


	/* Handle by grab type, assume error first. */
	last_grab_error = "Specified grab operation no supported";
	status = -1;
	switch(grab_data->grab_type)
	{
	  case GRAB_WINDOW:
#ifdef HAVE_X
	    status = grab_X_window(
		iv, grab_data, width, height,
		include_decorations, resize_on_load
	    );
#endif
	    break;

	  case GRAB_RECTANGULAR:
#ifdef HAVE_X
            status = grab_X_rectangular(
                iv, grab_data, width, height,
                include_decorations, resize_on_load
            );
#endif
	    break;

	  case GRAB_ABORT:
            last_grab_error = "User aborted";
            GRAB_CLEANUP
            reenterent = FALSE;
            return(-4);
	    break;

	  default:
	    last_grab_error = "Grab has been aborted";
	    break;
	}

	/* Ungrab pointer and disconnect signal handlers. */
	GRAB_CLEANUP
#undef GRAB_CLEANUP
#undef DO_UNGRAB_POINTER
#undef DO_FREE_GRAB_CBDATA

	reenterent = FALSE;
	return(0);
}
